(* TEST
include tool-ocaml-lib
flags = "-w -a"
ocaml_script_as_argument = "true"
* setup-ocaml-build-env
** ocaml
*)

1;;

(**
       0 CONST1
       1 ATOM0
       2 SETGLOBAL T010-const1
       4 STOP
**)
