(* TEST
include tool-ocaml-lib
flags = "-w -a"
ocaml_script_as_argument = "true"
* setup-ocaml-build-env
** ocaml
*)

2;;

(**
       0 CONST2
       1 ATOM0
       2 SETGLOBAL T010-const2
       4 STOP
**)
