(* TEST
* setup-ocamlc.byte-build-env
** ocamlc.byte
ocamlc_byte_exit_status = "2"
*** check-ocamlc.byte-output
*)

(* It is apparently impossible to get the "unclosed object" message. *)

class c : object
